@@ -16,13 +16,14 @@
 ######################################################################
 
 import sys     # for argv[]
-import os      # for popen2()
+import os
 import shutil  # for rmtree()
 import re
 import stat    # for ST_MODE
 import copy    # for deepcopy()
 import time    # for time()
 import traceback # for print_exc()
+import threading
 
 import getopt
 try:
@@ -58,12 +59,6 @@ from svntest import wc
 #####################################################################
 # Global stuff
 
-### Grandfather in SVNTreeUnequal, which used to live here.  If you're
-# ever feeling saucy, you could go through the testsuite and change
-# main.SVNTreeUnequal to test.SVNTreeUnequal.
-import tree
-SVNTreeUnequal = tree.SVNTreeUnequal
-
 class SVNProcessTerminatedBySignal(Failure):
   "Exception raised if a spawned process segfaulted, aborted, etc."
   pass
@@ -90,28 +85,20 @@ class SVNRepositoryCreateFailure(Failure):
 
 # Windows specifics
 if sys.platform == 'win32':
-  windows = 1
+  windows = True
   file_scheme_prefix = 'file:///'
   _exe = '.exe'
 else:
-  windows = 0
+  windows = False
   file_scheme_prefix = 'file://'
   _exe = ''
 
 # os.wait() specifics
 try:
   from os import wait
-  platform_with_os_wait = 1
+  platform_with_os_wait = True
 except ImportError:
-  platform_with_os_wait = 0
-
-# The locations of the svn, svnadmin and svnlook binaries, relative to
-# the only scripts that import this file right now (they live in ../).
-svn_binary = os.path.abspath('../../svn/svn' + _exe)
-svnadmin_binary = os.path.abspath('../../svnadmin/svnadmin' + _exe)
-svnlook_binary = os.path.abspath('../../svnlook/svnlook' + _exe)
-svnsync_binary = os.path.abspath('../../svnsync/svnsync' + _exe)
-svnversion_binary = os.path.abspath('../../svnversion/svnversion' + _exe)
+  platform_with_os_wait = False
 
 # The location of our mock svneditor script.
 svneditor_script = os.path.join(sys.path[0], 'svneditor.py')
@@ -124,27 +111,71 @@ wc_passwd = 'rayjandom'
 # scenarios
 wc_author2 = 'jconstant' # use the same password as wc_author
 
-# Global variable indicating if we want verbose output.
-verbose_mode = 0
+######################################################################
+# Global variables set during option parsing.  These should not be used
+# until the variable command_line_parsed has been set to True, as is
+# done in run_tests below.
+command_line_parsed = False
+
+# The locations of the svn, svnadmin and svnlook binaries, relative to
+# the only scripts that import this file right now (they live in ../).
+# Use --bin to override these defaults.
+svn_binary = os.path.abspath('../../svn/svn' + _exe)
+svnadmin_binary = os.path.abspath('../../svnadmin/svnadmin' + _exe)
+svnlook_binary = os.path.abspath('../../svnlook/svnlook' + _exe)
+svnsync_binary = os.path.abspath('../../svnsync/svnsync' + _exe)
+svnversion_binary = os.path.abspath('../../svnversion/svnversion' + _exe)
+
+# Global variable indicating if we want verbose output, that is, 
+# details of what commands each test does as it does them.  This is
+# incompatible with quiet_mode.
+verbose_mode = False
+
+# Global variable indicating if we want quiet output, that is, don't
+# show PASS, XFAIL, or SKIP notices, but do show FAIL and XPASS.  This
+# is incompatible with verbose_mode.
+quiet_mode = False
 
 # Global variable indicating if we want test data cleaned up after success
-cleanup_mode = 0
+cleanup_mode = False
 
 # Global variable indicating if svnserve should use Cyrus SASL
-enable_sasl = 0
+enable_sasl = False
+
+# Global variable indicating which DAV library, if any, is in use
+# ('neon', 'serf')
+http_library = None
+
+# Global variable indicating what the minor version of the server
+# tested against is (4 for 1.4.x, for example).
+server_minor_version = 5
+
+# Global variable indicating if this is a child process and no cleanup
+# of global directories is needed.
+is_child_process = False
 
 # Global URL to testing area.  Default to ra_local, current working dir.
 test_area_url = file_scheme_prefix + os.path.abspath(os.getcwd())
-if windows == 1:
+if windows:
   test_area_url = test_area_url.replace('\\', '/')
 
+# Location to the pristine repository, will be calculated from test_area_url
+# when we know what the user specified for --url.
+pristine_url = None
+
 # Global variable indicating the FS type for repository creations.
 fs_type = None
 
+# End of command-line-set global variables.
+######################################################################
+
 # All temporary repositories and working copies are created underneath
 # this dir, so there's one point at which to mount, e.g., a ramdisk.
 work_dir = "svn-test-work"
 
+# Constant for the merge info property.
+SVN_PROP_MERGE_INFO = "svn:mergeinfo"
+
 # Where we want all the repositories and working copies to live.
 # Each test will have its own!
 general_repo_dir = os.path.join(work_dir, "repositories")
@@ -160,10 +191,6 @@ pristine_dir = os.path.join(temp_dir, "repos")
 greek_dump_dir = os.path.join(temp_dir, "greekfiles")
 default_config_dir = os.path.abspath(os.path.join(temp_dir, "config"))
 
-# Location to the pristine repository, will be calculated from test_area_url
-# when we know what the user specified for --url.
-pristine_url = None
-
 #
 # Our pristine greek-tree state.
 #
@@ -243,35 +270,46 @@ def run_command(command, error_expected, binary_mode=0, *varargs):
   return run_command_stdin(command, error_expected, binary_mode,
                            None, *varargs)
 
-# Run any binary, supplying input text, logging the command line
-def run_command_stdin(command, error_expected, binary_mode=0,
-                      stdin_lines=None, *varargs):
-  """Run COMMAND with VARARGS; input STDIN_LINES (a list of strings
-  which should include newline characters) to program via stdin - this
-  should not be very large, as if the program outputs more than the OS
-  is willing to buffer, this will deadlock, with both Python and
-  COMMAND waiting to write to each other for ever.
-  Return stdout, stderr as lists of lines.
-  If ERROR_EXPECTED is None, any stderr also will be printed."""
+# A regular expression that matches arguments that are trivially safe
+# to pass on a command line without quoting on any supported operating
+# system:
+_safe_arg_re = re.compile(r'^[A-Za-z\d\.\_\/\-\:\@]+$')
 
-  args = ''
-  for arg in varargs:                   # build the command string
-    arg = str(arg)
+def _quote_arg(arg):
+  """Quote ARG for a command line.
+
+  Simply surround every argument in double-quotes unless it contains
+  only universally harmless characters.
+
+  WARNING: This function cannot handle arbitrary command-line
+  arguments.  It can easily be confused by shell metacharacters.  A
+  perfect job would be difficult and OS-dependent (see, for example,
+  http://msdn.microsoft.com/library/en-us/vccelng/htm/progs_12.asp).
+  In other words, this function is just good enough for what we need
+  here."""
+
+  arg = str(arg)
+  if _safe_arg_re.match(arg):
+    return arg
+  else:
     if os.name != 'nt':
       arg = arg.replace('$', '\$')
-    args = args + ' "' + arg + '"'
+    return '"%s"' % (arg,)
+
+# Run any binary, supplying input text, logging the command line
+def spawn_process(command, binary_mode=0,stdin_lines=None, *varargs):
+  args = ' '.join(map(_quote_arg, varargs))
 
   # Log the command line
   if verbose_mode:
-    print 'CMD:', os.path.basename(command) + args,
+    print 'CMD:', os.path.basename(command) + ' ' + args,
 
   if binary_mode:
     mode = 'b'
   else:
     mode = 't'
 
-  start = time.time()
-  infile, outfile, errfile = os.popen3(command + args, mode)
+  infile, outfile, errfile = os.popen3(command + ' ' + args, mode)
 
   if stdin_lines:
     map(infile.write, stdin_lines)
@@ -284,16 +322,44 @@ def run_command_stdin(command, error_expected, binary_mode=0,
   outfile.close()
   errfile.close()
 
+  exit_code = 0
+
   if platform_with_os_wait:
     pid, wait_code = os.wait()
 
-    exit_code = int(wait_code / 256)
-    exit_signal = wait_code % 256
-
-    if exit_signal != 0:
+    if os.WIFSIGNALED(wait_code):
+      exit_signal = os.WTERMSIG(wait_code)
       sys.stdout.write("".join(stdout_lines))
       sys.stderr.write("".join(stderr_lines))
+      if verbose_mode:
+        # show the whole path to make it easier to start a debugger
+        sys.stderr.write("CMD: %s terminated by signal %d\n"
+                         % (command, exit_signal))
       raise SVNProcessTerminatedBySignal
+    else:
+      exit_code = os.WEXITSTATUS(wait_code)
+      if exit_code and verbose_mode:
+        sys.stderr.write("CMD: %s exited with %d\n" % (command, exit_code))
+
+  return exit_code, stdout_lines, stderr_lines
+
+def run_command_stdin(command, error_expected, binary_mode=0,
+                      stdin_lines=None, *varargs):
+  """Run COMMAND with VARARGS; input STDIN_LINES (a list of strings
+  which should include newline characters) to program via stdin - this
+  should not be very large, as if the program outputs more than the OS
+  is willing to buffer, this will deadlock, with both Python and
+  COMMAND waiting to write to each other for ever.
+  Return stdout, stderr as lists of lines.
+  If ERROR_EXPECTED is None, any stderr also will be printed."""
+
+  if verbose_mode:
+    start = time.time()
+
+  exit_code, stdout_lines, stderr_lines = spawn_process(command,
+                                                        binary_mode,
+                                                        stdin_lines,
+                                                        *varargs)
 
   if verbose_mode:
     stop = time.time()
@@ -305,22 +371,44 @@ def run_command_stdin(command, error_expected, binary_mode=0,
 
   return stdout_lines, stderr_lines
 
-def create_config_dir(cfgdir,
-                      config_contents = '#\n',
-                      server_contents = '#\n'):
+def create_config_dir(cfgdir, config_contents=None, server_contents=None):
   "Create config directories and files"
 
   # config file names
   cfgfile_cfg = os.path.join(cfgdir, 'config')
-  cfgfile_srv = os.path.join(cfgdir, 'server')
+  cfgfile_srv = os.path.join(cfgdir, 'servers')
 
   # create the directory
   if not os.path.isdir(cfgdir):
     os.makedirs(cfgdir)
 
+  # define default config file contents if none provided
+  if config_contents is None:
+    config_contents = """
+#
+[miscellany]
+interactive-conflicts = false
+"""
+
+  # define default server file contents if none provided
+  if server_contents is None:
+    if http_library:
+      server_contents = """
+#
+[global]
+http-library=%s
+""" % (http_library)
+    else:
+      server_contents = "#\n"
+
   file_write(cfgfile_cfg, config_contents)
   file_write(cfgfile_srv, server_contents)
 
+def _with_config_dir(args):
+  if '--config-dir' in args:
+    return args
+  else:
+    return args + ('--config-dir', default_config_dir)
 
 # For running subversion and returning the output
 def run_svn(error_expected, *varargs):
@@ -328,12 +416,7 @@ def run_svn(error_expected, *varargs):
   If ERROR_EXPECTED is None, any stderr also will be printed.  If
   you're just checking that something does/doesn't come out of
   stdout/stderr, you might want to use actions.run_and_verify_svn()."""
-  if '--config-dir' in varargs:
-    return run_command(svn_binary, error_expected, 0,
-                       *varargs)
-  else:
-    return run_command(svn_binary, error_expected, 0,
-                       *varargs + ('--config-dir', default_config_dir))
+  return run_command(svn_binary, error_expected, 0, *(_with_config_dir(varargs)))
 
 # For running svnadmin.  Ignores the output.
 def run_svnadmin(*varargs):
@@ -347,7 +430,7 @@ def run_svnlook(*varargs):
 
 def run_svnsync(*varargs):
   "Run svnsync with VARARGS, returns stdout, stderr as list of lines."
-  return run_command(svnsync_binary, 1, 0, *varargs)
+  return run_command(svnsync_binary, 1, 0, *(_with_config_dir(varargs)))
 
 def run_svnversion(*varargs):
   "Run svnversion with VARARGS, returns stdout, stderr as list of lines."
@@ -395,7 +478,7 @@ def file_append(path, new_text):
 def file_append_binary(path, new_text):
   "Append NEW_TEXT to file at PATH in binary mode"
   file_write(path, new_text, 'ab')  # open in (a)ppend mode
-  
+
 # For creating new files, and making local mods to existing files.
 def file_write(path, contents, mode = 'w'):
   """Write the CONTENTS to the file at PATH, opening file using MODE,
@@ -404,6 +487,26 @@ def file_write(path, contents, mode = 'w'):
   fp.write(contents)
   fp.close()
 
+# For reading the contents of a file
+def file_read(path, mode = 'r'):
+  """Return the contents of the file at PATH, opening file using MODE,
+  which is (r)ead by default."""
+  fp = open(path, mode)
+  contents = fp.read()
+  fp.close()
+  return contents
+
+# For replacing parts of contents in an existing file, with new content.
+def file_substitute(path, contents, new_contents):
+  """Replace the CONTENTS in the file at PATH using the NEW_CONTENTS"""
+  fp = open(path, 'r')
+  fcontent = fp.read()
+  fp.close()
+  fcontent = fcontent.replace(contents, new_contents)
+  fp = open(path, 'w')
+  fp.write(fcontent)
+  fp.close()
+
 # For creating blank new repositories
 def create_repos(path):
   """Create a brand-new SVN repository at PATH.  If PATH does not yet
@@ -413,6 +516,8 @@ def create_repos(path):
     os.makedirs(path) # this creates all the intermediate dirs, if neccessary
 
   opts = ("--bdb-txn-nosync",)
+  if server_minor_version < 5:
+    opts += ("--pre-1.5-compatible",)
   if fs_type is not None:
     opts += ("--fs-type=" + fs_type,)
   stdout, stderr = run_command(svnadmin_binary, 1, 0, "create", path, *opts)
@@ -429,7 +534,7 @@ def create_repos(path):
   # Allow unauthenticated users to write to the repos, for ra_svn testing.
   file_write(get_svnserve_conf_file_path(path),
              "[general]\nauth-access = write\n");
-  if enable_sasl == 1:
+  if enable_sasl:
     file_append(get_svnserve_conf_file_path(path),
                 "realm = svntest\n[sasl]\nuse-sasl = true\n")
   else:
@@ -440,10 +545,10 @@ def create_repos(path):
   chmod_tree(path, 0666, 0666)
 
 # For copying a repository
-def copy_repos(src_path, dst_path, head_revision, ignore_uuid = 0):
+def copy_repos(src_path, dst_path, head_revision, ignore_uuid = 1):
   "Copy the repository SRC_PATH, with head revision HEAD_REVISION, to DST_PATH"
 
-  # Do an svnadmin dump|svnadmin load cycle. Print a fake pipe command so that 
+  # Do an svnadmin dump|svnadmin load cycle. Print a fake pipe command so that
   # the displayed CMDs can be run by hand
   create_repos(dst_path)
   dump_args = ' dump "' + src_path + '"'
@@ -521,50 +626,116 @@ def create_python_hook_script (hook_path, hook_script_code):
     # Use an absolute path since the working directory is not guaranteed
     hook_path = os.path.abspath(hook_path)
     # Fill the python file.
-    file_append ("%s.py" % hook_path, hook_script_code)
+    file_write ("%s.py" % hook_path, hook_script_code)
     # Fill the batch wrapper file.
     file_append ("%s.bat" % hook_path,
-                 "@\"%s\" %s.py\n" % (sys.executable, hook_path))
+                 "@\"%s\" %s.py %%*\n" % (sys.executable, hook_path))
   else:
     # For all other platforms
-    file_append (hook_path, "#!%s\n%s" % (sys.executable, hook_script_code))
+    file_write (hook_path, "#!%s\n%s" % (sys.executable, hook_script_code))
     os.chmod (hook_path, 0755)
 
+def write_restrictive_svnserve_conf(repo_dir, anon_access="none"):
+  "Create a restrictive authz file ( no anynomous access )."
 
-def compare_unordered_output(expected, actual):
-  """Compare lists of output lines for equality disregarding the
-     order of the lines"""
-  if len(actual) != len(expected):
-    raise Failure("Length of expected output not equal to actual length")
+  fp = open(get_svnserve_conf_file_path(repo_dir), 'w')
+  fp.write("[general]\nanon-access = %s\nauth-access = write\n"
+           "authz-db = authz\n" % anon_access)
+  if enable_sasl == 1:
+    fp.write("realm = svntest\n[sasl]\nuse-sasl = true\n");
+  else:
+    fp.write("password-db = passwd\n")
+  fp.close()
 
-  expected = list(expected)
-  for aline in actual:
-    try:
-      i = expected.index(aline)
-      expected.pop(i)
-    except ValueError:
-      raise Failure("Expected output does not match actual output")
+def write_authz_file(sbox, rules, sections=None):
+  """Write an authz file to SBOX, appropriate for the RA method used,
+with authorizations rules RULES mapping paths to strings containing
+the rules. You can add sections SECTIONS (ex. groups, aliases...) with
+an appropriate list of mappings.
+"""
+  fp = open(sbox.authz_file, 'w')
+  if sbox.repo_url.startswith("http"):
+    prefix = sbox.name + ":"
+  else:
+    prefix = ""
+  if sections:
+    for p, r in sections.items():
+      fp.write("[%s]\n%s\n" % (p, r))
+
+  for p, r in rules.items():
+    fp.write("[%s%s]\n%s\n" % (prefix, p, r))
+  fp.close()
 
 def use_editor(func):
   os.environ['SVN_EDITOR'] = svneditor_script
+  os.environ['SVN_MERGE'] = svneditor_script
   os.environ['SVNTEST_EDITOR_FUNC'] = func
 
+
+def merge_notify_line(revstart=None, revend=None):
+  """Return an expected output line that describes the beginning of a
+  merge operation on revisions REVSTART through REVEND.  Omit both
+  REVSTART and REVEND for the case where the left and right sides of
+  the merge are from different URLs."""
+  if revend is None:
+    if revstart is None:
+      # The left and right sides of the merge are from different URLs.
+      return "--- Merging differences between repository URLs into '.+':\n"
+    elif revstart < 0:
+      return "--- Reverse-merging r%ld into '.+':\n" % abs(revstart)
+    else:
+      return "--- Merging r%ld into '.+':\n" % revstart
+  else:
+    if revstart > revend:
+      return "--- Reverse-merging r%ld through r%ld into '.+':\n" % (revstart,
+                                                                     revend)
+    else:
+      return "--- Merging r%ld through r%ld into '.+':\n" % (revstart, revend)
+
+
 ######################################################################
 # Functions which check the test configuration
 # (useful for conditional XFails)
 
+def _check_command_line_parsed():
+  """Raise an exception if the command line has not yet been parsed."""
+  if not command_line_parsed:
+    raise Failure("Condition cannot be tested until command line is parsed")
+
 def is_ra_type_dav():
+  _check_command_line_parsed()
   return test_area_url.startswith('http')
 
 def is_ra_type_svn():
+  _check_command_line_parsed()
   return test_area_url.startswith('svn')
 
+def is_ra_type_file():
+  _check_command_line_parsed()
+  return test_area_url.startswith('file')
+
 def is_fs_type_fsfs():
+  _check_command_line_parsed()
   # This assumes that fsfs is the default fs implementation.
-  return (fs_type == 'fsfs' or fs_type is None)
+  return fs_type == 'fsfs' or fs_type is None
 
 def is_os_windows():
-  return (os.name == 'nt')
+  return os.name == 'nt'
+
+def is_posix_os():
+  return os.name == 'posix'
+
+def is_os_darwin():
+  return sys.platform == 'darwin'
+
+def server_has_mergeinfo():
+  _check_command_line_parsed()
+  return server_minor_version >= 5
+
+def server_has_revprop_commit():
+  _check_command_line_parsed()
+  return server_minor_version >= 5
+
 
 ######################################################################
 # Sandbox handling
@@ -602,7 +773,7 @@ class Sandbox:
     elif self.repo_url.startswith("svn"):
       self.authz_file = os.path.join(self.repo_dir, "conf", "authz")
 
-    if windows == 1:
+    if windows:
       self.repo_url = self.repo_url.replace('\\', '/')
     self.test_paths = [self.wc_dir, self.repo_dir]
 
@@ -670,6 +841,44 @@ def _cleanup_test_path(path, retrying=None):
       print "WARNING: cleanup failed, will try again later"
     _deferred_test_paths.append(path)
 
+class SpawnTest(threading.Thread):
+  """Encapsulate a single test case, run it in a separate child process.
+  Instead of waiting till the process is finished, add this class to a
+  list of active tests for follow up in the parent process."""
+  def __init__(self, index, tests = None):
+    threading.Thread.__init__(self)
+    self.index = index
+    self.tests = tests
+    self.result = None
+    self.stdout_lines = None
+    self.stderr_lines = None
+
+  def run(self):
+    command = sys.argv[0]
+
+    args = []
+    args.append(str(self.index))
+    args.append('-c')
+    # add some startup arguments from this process
+    if fs_type:
+      args.append('--fs-type=' + fs_type)
+    if test_area_url:
+      args.append('--url=' + test_area_url)
+    if verbose_mode:
+      args.append('-v')
+    if cleanup_mode:
+      args.append('--cleanup')
+    if enable_sasl:
+      args.append('--enable-sasl')
+
+    self.result, self.stdout_lines, self.stderr_lines =\
+                                         spawn_process(command, 1, None, *args)
+    # don't trust the exitcode, will not be correct on Windows
+    if filter(lambda x: x.startswith('FAIL: ') or x.startswith('XPASS: '),
+              self.stdout_lines):
+      self.result = 1
+    self.tests.append(self)
+    sys.stdout.write('.')
 
 class TestRunner:
   """Encapsulate a single test case (predicate), including logic for
@@ -709,7 +918,9 @@ class TestRunner:
     # Tests that want to use an editor should invoke svntest.main.use_editor.
     os.environ['SVN_EDITOR'] = ''
     os.environ['SVNTEST_EDITOR_FUNC'] = ''
+    actions.no_sleep_for_timestamps()
 
+    saved_dir = os.getcwd()
     try:
       rc = apply(self.pred.run, (), kw)
       if rc is not None:
@@ -732,6 +943,7 @@ class TestRunner:
           print 'EXCEPTION: %s: %s' % (ex.__class__.__name__, ex_args)
         else:
           print 'EXCEPTION:', ex.__class__.__name__
+      traceback.print_exc(file=sys.stdout)
     except KeyboardInterrupt:
       print 'Interrupted'
       sys.exit(0)
@@ -744,15 +956,18 @@ class TestRunner:
       result = 1
       print 'UNEXPECTED EXCEPTION:'
       traceback.print_exc(file=sys.stdout)
+
+    os.chdir(saved_dir)
     result = self.pred.convert_result(result)
-    print self.pred.run_text(result),
-    self._print_name()
-    sys.stdout.flush()
+    (result_text, result_benignity) = self.pred.run_text(result)
+    if not (quiet_mode and result_benignity):
+      print result_text,
+      self._print_name()
+      sys.stdout.flush()
     if sandbox is not None and result != 1 and cleanup_mode:
       sandbox.cleanup_test_paths()
     return result
 
-
 ######################################################################
 # Main testing functions
 
@@ -763,26 +978,71 @@ class TestRunner:
 # it can be displayed by the 'list' command.)
 
 # Func to run one test in the list.
-def run_one_test(n, test_list):
-  "Run the Nth client test in TEST_LIST, return the result."
+def run_one_test(n, test_list, parallel = 0, finished_tests = None):
+  """Run the Nth client test in TEST_LIST, return the result.
+
+  If we're running the tests in parallel spawn the test in a new process.
+  """
 
   if (n < 1) or (n > len(test_list) - 1):
     print "There is no test", `n` + ".\n"
     return 1
 
   # Run the test.
-  exit_code = TestRunner(test_list[n], n).run()
-  return exit_code
+  if parallel:
+    st = SpawnTest(n, finished_tests)
+    st.start()
+    return 0
+  else:
+    exit_code = TestRunner(test_list[n], n).run()
+    return exit_code
 
-def _internal_run_tests(test_list, testnums):
-  """Run the tests from TEST_LIST whose indices are listed in TESTNUMS."""
+def _internal_run_tests(test_list, testnums, parallel):
+  """Run the tests from TEST_LIST whose indices are listed in TESTNUMS.
+
+  If we're running the tests in parallel spawn as much parallel processes
+  as requested and gather the results in a temp. buffer when a child
+  process is finished.
+  """
 
   exit_code = 0
+  finished_tests = []
+  tests_started = 0
 
-  for testnum in testnums:
-    # 1 is the only return code that indicates actual test failure.
-    if run_one_test(testnum, test_list) == 1:
-      exit_code = 1
+  if not parallel:
+    for testnum in testnums:
+      if run_one_test(testnum, test_list) == 1:
+          exit_code = 1
+  else:
+    for testnum in testnums:
+      # wait till there's a free spot.
+      while tests_started - len(finished_tests) > parallel:
+        time.sleep(0.2)
+      run_one_test(testnum, test_list, parallel, finished_tests)
+      tests_started += 1
+
+    # wait for all tests to finish
+    while len(finished_tests) < len(testnums):
+      time.sleep(0.2)
+
+    # Sort test results list by test nr.
+    deco = [(test.index, test) for test in finished_tests]
+    deco.sort()
+    finished_tests = [test for (ti, test) in deco]
+
+    # terminate the line of dots
+    print
+
+    # all tests are finished, find out the result and print the logs.
+    for test in finished_tests:
+      if test.stdout_lines:
+        for line in test.stdout_lines:
+          sys.stdout.write(line)
+      if test.stderr_lines:
+        for line in test.stderr_lines:
+          sys.stdout.write(line)
+      if test.result == 1:
+        exit_code = 1
 
   _cleanup_deferred_test_paths()
   return exit_code
@@ -790,28 +1050,33 @@ def _internal_run_tests(test_list, testnums):
 
 def usage():
   prog_name = os.path.basename(sys.argv[0])
-  print "%s [--url] [--fs-type] [--verbose] [--enable-sasl] [--cleanup] \\" \
-        % prog_name
-  print "%s [<test> ...]" % (" " * len(prog_name))
+  print "%s [--url] [--fs-type] [--verbose|--quiet] \\" % prog_name
+  print "%s [--enable-sasl] [--cleanup] [--bin] [<test> ...]" \
+      % (" " * len(prog_name))
+  print "%s " % (" " * len(prog_name))
   print "%s [--list] [<test> ...]\n" % prog_name
   print "Arguments:"
   print " test          The number of the test to run (multiple okay), " \
         "or all tests\n"
   print "Options:"
-  print " --list        Print test doc strings instead of running them"
-  print " --fs-type     Subversion file system type (fsfs or bdb)"
-  print " --url         Base url to the repos (e.g. svn://localhost)"
-  print " --verbose     Print binary command-lines"
-  print " --cleanup     Whether to clean up"
-  print " --enable-sasl Whether to enable SASL authentication"
-  print " --help        This information"
+  print " --list          Print test doc strings instead of running them"
+  print " --fs-type       Subversion file system type (fsfs or bdb)"
+  print " --http-library  DAV library to use (neon or serf)"
+  print " --url           Base url to the repos (e.g. svn://localhost)"
+  print " --verbose       Print binary command-lines (not with --quiet)"
+  print " --quiet         Print only unexpected results (not with --verbose)"
+  print " --cleanup       Whether to clean up"
+  print " --enable-sasl   Whether to enable SASL authentication"
+  print " --parallel      Run the tests in parallel"
+  print " --bin           Use the svn binaries installed in this path"
+  print " --help          This information"
 
 
 # Main func.  This is the "entry point" that all the test scripts call
 # to run their list of tests.
 #
 # This routine parses sys.argv to decide what to do.
-def run_tests(test_list):
+def run_tests(test_list, serial_only = False):
   """Main routine to run all tests in TEST_LIST.
 
   NOTE: this function does not return. It does a sys.exit() with the
@@ -822,20 +1087,39 @@ def run_tests(test_list):
   global pristine_url
   global fs_type
   global verbose_mode
+  global quiet_mode
   global cleanup_mode
   global enable_sasl
+  global is_child_process
+  global svn_binary
+  global svnadmin_binary
+  global svnlook_binary
+  global svnsync_binary
+  global svnversion_binary
+  global command_line_parsed
+  global http_library
+  global server_minor_version
+
   testnums = []
   # Should the tests be listed (as opposed to executed)?
-  list_tests = 0
+  list_tests = False
 
-  opts, args = my_getopt(sys.argv[1:], 'vh',
-                         ['url=', 'fs-type=', 'verbose', 'cleanup', 'list',
-                          'enable-sasl', 'help'])
+  parallel = 0
+  svn_bin = None
+  try:
+    opts, args = my_getopt(sys.argv[1:], 'vqhpc',
+                           ['url=', 'fs-type=', 'verbose', 'quiet', 'cleanup',
+                            'list', 'enable-sasl', 'help', 'parallel',
+                            'bin=', 'http-library=', 'server-minor-version='])
+  except getopt.GetoptError, e:
+    print "ERROR: %s\n" % e
+    usage()
+    sys.exit(1)
 
   for arg in args:
     if arg == "list":
       # This is an old deprecated variant of the "--list" option:
-      list_tests = 1
+      list_tests = True
     elif arg.startswith('BASE_URL='):
       test_area_url = arg[9:]
     else:
@@ -854,31 +1138,69 @@ def run_tests(test_list):
       fs_type = val
 
     elif opt == "-v" or opt == "--verbose":
-      verbose_mode = 1
+      verbose_mode = True
+
+    elif opt == "-q" or opt == "--quiet":
+      quiet_mode = True
 
     elif opt == "--cleanup":
-      cleanup_mode = 1
+      cleanup_mode = True
 
     elif opt == "--list":
-      list_tests = 1
+      list_tests = True
 
     elif opt == "--enable-sasl":
-      enable_sasl = 1
+      enable_sasl = True
 
     elif opt == "-h" or opt == "--help":
       usage()
       sys.exit(0)
 
+    elif opt == '-p' or opt == "--parallel":
+      parallel = 5   # use 5 parallel threads.
+
+    elif opt == '-c':
+      is_child_process = True
+
+    elif opt == '--bin':
+      svn_bin = val
+
+    elif opt == '--http-library':
+      http_library = val
+
+    elif opt == '--server-minor-version':
+      server_minor_version = int(val)
+      if server_minor_version < 4 or server_minor_version > 6:
+        print "ERROR: test harness only supports server minor version 4 or 5"
+        sys.exit(1)
+
   if test_area_url[-1:] == '/': # Normalize url to have no trailing slash
     test_area_url = test_area_url[:-1]
 
+  if verbose_mode and quiet_mode:
+    sys.stderr.write("ERROR: 'verbose' and 'quiet' are incompatible\n")
+    sys.exit(1)
+
   # Calculate pristine_url from test_area_url.
   pristine_url = test_area_url + '/' + pristine_dir
-  if windows == 1:
-    pristine_url = pristine_url.replace('\\', '/')  
-  
-  # Setup the pristine repository (and working copy)
-  actions.setup_pristine_repository()
+  if windows:
+    pristine_url = pristine_url.replace('\\', '/')
+
+  if not svn_bin is None:
+    svn_binary = os.path.join(svn_bin, 'svn' + _exe)
+    svnadmin_binary = os.path.join(svn_bin, 'svnadmin' + _exe)
+    svnlook_binary = os.path.join(svn_bin, 'svnlook' + _exe)
+    svnsync_binary = os.path.join(svn_bin, 'svnsync' + _exe)
+    svnversion_binary = os.path.join(svn_bin, 'svnversion' + _exe)
+
+  command_line_parsed = True
+
+  ######################################################################
+
+  # Cleanup: if a previous run crashed or interrupted the python
+  # interpreter, then `temp_dir' was never removed.  This can cause wonkiness.
+  if not is_child_process:
+    safe_rmtree(temp_dir)
 
   if not testnums:
     # If no test numbers were listed explicitly, include all of them:
@@ -893,26 +1215,30 @@ def run_tests(test_list):
     # done. just exit with success.
     sys.exit(0)
 
-  else:
-    exit_code = _internal_run_tests(test_list, testnums)
+  # don't run tests in parallel when the tests don't support it or there
+  # are only a few tests to run.
+  if serial_only or len(testnums) < 2:
+    parallel = 0
 
-    # remove all scratchwork: the 'pristine' repository, greek tree, etc.
-    # This ensures that an 'import' will happen the next time we run.
+  # Setup the pristine repository
+  actions.setup_pristine_repository()
+
+  # Build out the default configuration directory
+  create_config_dir(default_config_dir)
+
+  # Run the tests.
+  exit_code = _internal_run_tests(test_list, testnums, parallel)
+
+  # Remove all scratchwork: the 'pristine' repository, greek tree, etc.
+  # This ensures that an 'import' will happen the next time we run.
+  if not is_child_process:
     safe_rmtree(temp_dir)
 
-    _cleanup_deferred_test_paths()
+  # Cleanup after ourselves.
+  _cleanup_deferred_test_paths()
 
-    # return the appropriate exit code from the tests.
-    sys.exit(exit_code)
-
-
-######################################################################
-# Initialization
-
-# Cleanup: if a previous run crashed or interrupted the python
-# interpreter, then `temp_dir' was never removed.  This can cause wonkiness.
-
-safe_rmtree(temp_dir)
+  # Return the appropriate exit code from the tests.
+  sys.exit(exit_code)
 
 # the modules import each other, so we do this import very late, to ensure
 # that the definitions in "main" have been completed.
